import os

from cachetools import cached, LRUCache
from logging_helper import setup_logging

from ._auth import _get_token
from ._cache import get_file_cache
from ._compat import urlsplit
from ._github import check_repo

logger = setup_logging()

session = get_file_cache("rtd")

TOKEN = os.environ.get("RTD_AUTH_TOKEN")
logger.debug("TOKEN: {}".format(TOKEN))
headers = {}
if TOKEN:
    headers["Authorization"] = "Token {}".format(TOKEN)


class AuthenticationError(Exception):
    pass


class ReadtheDocs:
    def __init__(self, base_url="https://readthedocs.org/api/v3/"):
        self.headers = headers
        self.base_url = base_url
        token = _get_token("readthedocs.io")
        if token:
            self.headers["Authorization"] = "Token {}".format(TOKEN)

    def base(self):
        """Simple query just to check connection"""
        return session.get(self.base_url)

    @cached(cache=LRUCache(maxsize=12))
    def get_project(self, name, version=None):
        if "/v2/" in self.base_url:
            url = "{}footer_html/?format=json&project={}&version=v{}".format(
                name, version
            )
        else:
            url = "{}projects/{}".format(self.base_url, name)
        logger.debug("rtd request: {}".format(url))
        response = session.get(url, headers=self.headers)
        logger.debug(response.headers)
        try:
            data = response.json()
        except Exception as e:
            logger.error("get_project: {}".format(e))
            return

        if data == {"detail": "Invalid token."}:
            raise AuthenticationError("Invalid token")

        logger.debug("rtd {}: {}".format(name, data))

        repo = data.get("repository")
        logger.info(repo)

        if not repo:
            return

        url = repo.get("url")
        return url


@cached(cache=LRUCache(maxsize=12))
def _get_repo_v2_footer(name, version="latest", dot_com=True):
    url = "https://readthedocs.{}/api/v2/footer_html/?format=json&project={}&version={}&page=index&docroot=/&source_suffix=.rst"
    url = url.format("com" if dot_com else "org", name, version)

    response = session.get(url)
    logger.debug(
        "rtd v2: {} {} {} {}".format(name, response, response.headers, response.content)
    )
    try:
        data = response.json()
    except Exception as e:
        logger.error("rtd v2: {}".format(e))
        return

    html = data.get("html")
    if not html:
        return

    pos = html.find('"https://github.com/')
    if pos == -1:
        pos = html.find('"https://bitbucket.org')
    if pos == -1:
        return
    html = html[pos + 1 : pos + 100]
    end = html.find("/blob/")
    if end == -1:
        end = html.find("/src/")
    url = html[:end]
    return url


def get_repo(slug, version="latest", dot_com=None, v2=None, strip_docs_suffix=True):
    if "://" in slug:
        p = urlsplit(slug)
        path = p.path
        if path:
            path = p.path[1:]

        host_parts = p.netloc.split(".")
        if len(host_parts) == 3:
            subdomain, _, hostname = p.netloc.partition(".")
            if subdomain in ["docs"]:
                return
            if hostname == "readthedocs-hosted.com":
                dot_com = True
            if path:
                try:
                    lang, version, _ = path.split("/", 2)
                except ValueError:
                    logger.info("rtd Couldnt split {}".format(path))
        else:
            assert len(host_parts) == 2, "{} unknown".format(host_parts)
            assert host_parts[0] == "readthedocs"
            assert path.startswith("projects/"), "{} unknown".format(path)
            try:
                _, subdomain, _ = path.split("/", 2)
            except ValueError:
                logger.info("rtd Couldnt split {}".format(path))
    else:
        subdomain = slug

    if subdomain in ["docs", "www", "assets"]:
        return

    if subdomain in ["python-jsonschema-objects"]:
        v2 = True

    url = None
    if not dot_com and not v2:
        rtd = ReadtheDocs()
        try:
            url = rtd.get_project(subdomain)
        except AuthenticationError:
            pass

    if not url:
        url = _get_repo_v2_footer(subdomain, version, dot_com)

    if not url:
        return

    if url.endswith(".git"):
        url = url[:-4]

    if url.startswith("git://"):
        url = "https://{}".format(url[6:])
    elif url.startswith("http://"):
        url = "https://github.com" + url[len("http://github.com") :]

    if url == "https://github.com/zzzeek/redirectthedocs":
        return

    if strip_docs_suffix:
        docless_url = None
        if url.endswith("-docs"):
            docless_url = url[:-5]
        elif url.endswith("doc"):
            docless_url = url[:-3]
        elif url.endswith("docs"):
            docless_url = url[:-4]
        elif url.endswith("/docs"):
            docless_url = url[:-5]
        elif url.endswith("/documentation"):
            docless_url = url[: len("/documentation")]
        elif url.endswith("/devguide"):  # CPython
            return
        elif url.endswith("/packaging.python.org"):
            return
        elif url.endswith("/pypa.io"):  # certifi
            return

        if docless_url and "github.com" in docless_url:
            if check_repo(docless_url):
                return docless_url

    return url



ALLOWED_STDLIB_BACKPORTS = set(
    [
        "argparse",
        "asyncio",
        "contextvars",
        "configparser",
        "csv",
        "ctypes",
        "dataclasses",
        "datetime",  # Zope, installs to non-conflicting DateTime on case sensitive fs
        "distutils",
        "email",
        "enum",
        "faulthandler",
        "functools",
        "hmac",
        "http",
        "functools",
        "futures",
        "hashlib",
        "html",
        "importlib",
        "ipaddress",
        "logging",
        "multiprocessing",
        "pathlib",
        "readline",
        "selectors",  # recheck
        "ssl",
        "statistics",
        "typing",
        "unittest",
        "uuid",
        "wsgiref",
    ]
)



class Input(object):
    def __init__(self, value, source=None):
        self._value = value
        self.source = source

    def __str__(self):  # pragma: no cover
        return self.value + " @ " + (self.source or "Unknown")

    def __repr__(self):
        if self.source:
            return "{}({!r}) @ {}".format(
                self.__class__.__name__, self.value, self.source
            )
        return "{}({!r})".format(self.__class__.__name__, self.value)

    @property
    def value(self):
        return self._value


class Url(Input):
    pass


class UrlSet(Input):
    def __str__(self):
        return "{!r}".format(self.value)


class Email(Input):
    pass


class Name(Input):
    pass


class Text(Input):
    def __str__(self):
        return self.value[:10] + "..."

    def __repr__(self):
        return "{}({!r})".format(self.__class__.__name__, str(self))


class Webpage(Text):
    def __str__(self):  # pragma: no cover
        return self.value[:10] + "..."

    @property
    def value(self):
        assert self._value
        assert self._value.text
        return self._value.text



#!/usr/bin/env python
"""PyPI client side database."""

"""
Copyright 2020 John Vandenberg

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from setuptools import find_packages, setup

__version__ = "0.2.4"

classifiers = """\
Environment :: Console
Environment :: Web Environment
Intended Audience :: Developers
Intended Audience :: Science/Research
Intended Audience :: System Administrators
License :: OSI Approved :: Apache Software License
Operating System :: OS Independent
Programming Language :: Python :: 2.7
Programming Language :: Python :: 3.5
Programming Language :: Python :: 3.6
Programming Language :: Python :: 3.7
Programming Language :: Python :: 3.8
Programming Language :: Python :: Implementation :: CPython
Development Status :: 3 - Alpha
"""

setup(
    name="pypidb",
    version=__version__,
    description="Local database of PyPI metadata to find VCS URLs",
    license="Apache-2.0",
    author_email="jayvdb@gmail.com",
    url="https://github.com/jayvdb/pypidb",
    packages=find_packages(exclude=["tests", "tests.*"]),
    package_data={"": ["*.txt"]},
    python_requires=">=2.7,!=3.0.*,!=3.1.*,!=3.2.*,!=3.3.*,!=3.4.*",
    install_requires=[
        "https-everywhere",
        "requests[security]",
        "brotlipy",  # urllib3 optional dep
        "CacheControl",
        "cachetools",
        "lockfile",
        "dns-cache",
        "diskcache",
        "fake-useragent",
        "appdirs",
        "click",
        "lxml",
        "lplight",
        "socialregexes",
        "stdlib-list",
        "urlextract",
        "textdistance",
        "unidiff",
        "logging-helper",
    ],
    classifiers=classifiers.splitlines(),
    entry_points='''
        [console_scripts]
        pypidb=pypidb.cli:cli
    ''',
    tests_require=["pytest-blockage", "unittest-expander"],
)



class InvalidPackage(ValueError):
    pass


class InvalidPackageVersion(InvalidPackage):
    pass


class UnrecognisedStdlibBackport(InvalidPackage):
    pass


class IncompletePackageMetadata(ValueError):
    pass


class PackageWithoutFiles(IncompletePackageMetadata):
    pass


class PackageWithoutUrls(IncompletePackageMetadata):
    pass



from unittest_expander import expand, foreach

from tests.data import exact, mismatch, missing_repos
from tests.utils import _TestBase, normalize, web_session

expected = {}
expected.update(exact)
expected.update(mismatch)


class TestSpecific(_TestBase):
    def test_odslib(self):
        url = self.converter.get_vcs("odslib")
        self.assertInsensitiveEqual(url, "https://bitbucket.org/angry_elf/odslib")

    def test_pytest_gae(self):
        url = self.converter.get_vcs("pytest-gae")
        self.assertInsensitiveEqual(url, "https://bitbucket.org/petraszd/pytest_gae")

    def test_hachoir_core(self):
        # https://bitbucket.org/haypo has been deleted
        url = self.converter.get_vcs("hachoir-core")
        self.assertInsensitiveEqual(url, "https://bitbucket.org/haypo/hachoir")

    def test_hachoir_metadata(self):
        url = self.converter.get_vcs("hachoir-metadata")
        self.assertInsensitiveEqual(url, "https://bitbucket.org/haypo/hachoir")

    def test_hachoir_parser(self):
        url = self.converter.get_vcs("hachoir-parser")
        self.assertInsensitiveEqual(url, "https://bitbucket.org/haypo/hachoir")


@expand
class TestExplicit(_TestBase):

    expected = expected
    names = [name for name, url in expected.items() if "bitbucket.org" in url]
    expected_failures = []

    @foreach(names)
    def test_package(self, name):
        expected = self.expected[name]
        try:
            url = self.converter.get_vcs(name)
        except Exception:
            if name in self.expected_failures:
                return
            raise

        if name in self.expected_failures:
            return

        self.assertIsNotNone(url)
        self.assertInsensitiveEqual(url, expected)

        normalised_name = normalize(name)

        if normalised_name in missing_repos:
            return
        r = web_session.get(url)
        r.raise_for_status()



from requests.exceptions import TooManyRedirects
from unittest_expander import expand, foreach

from pypidb._cache import get_file_cache
from pypidb._pypi import IncompletePackageMetadata, InvalidPackage, PackageWithoutUrls
from pypidb._similarity import _compute_similarity
from tests.data import (
    bad_metadata,
    missing_repos,
    setuppy_mismatches,
    sometimes_wrong_result,
    wrong_result,
)
from tests.test_rule_data import _rule_names
from tests.utils import _TestBase

web_session = get_file_cache("web")

failures = bad_metadata + sometimes_wrong_result + wrong_result + missing_repos

bad_metadata = [i for i in bad_metadata if i not in _rule_names]


@expand
class TestPyPiBadMetadata(_TestBase):
    @foreach(bad_metadata)
    def test_package(self, name):
        self.assertNotIn(name, wrong_result)
        self.assertNotIn(name, sometimes_wrong_result)
        self.assertNotIn(name, missing_repos)

        try:
            url = self._get_scm(name)
        except (InvalidPackage, IncompletePackageMetadata, PackageWithoutUrls):
            if self._debug_no_urls:
                self.assertTrue(False)
            url = None
        except TooManyRedirects:
            if self._debug_no_urls:
                self.assertTrue(False)
            url = None
        if self._debug_no_urls:
            self.assertIsNotNone(url)
        else:
            self.assertIsNone(url)


@expand
class TestWrong(_TestBase):
    @foreach(wrong_result)
    def test_package(self, name):
        self.assertNotIn(name, bad_metadata)
        self.assertNotIn(name, sometimes_wrong_result)
        self.assertNotIn(name, missing_repos)

        url = self._get_scm(name)
        self.assertIsNotNone(url)

        r = web_session.get(url, timeout=20)
        if r.status_code == 404:
            return
        r.raise_for_status()

        distance = _compute_similarity(name, url)
        self.assertGreater(
            distance, 0.13, "{} is a close match to {}".format(url, name)
        )


@expand
class TestWrongSometimes(_TestBase):
    @foreach(sometimes_wrong_result)
    def test_package(self, name):
        self.assertNotIn(name, bad_metadata)
        self.assertNotIn(name, wrong_result)
        self.assertNotIn(name, missing_repos)

        url = self._get_scm(name)
        if not url:
            return

        r = web_session.get(url, timeout=20)
        if r.status_code == 404:
            return
        r.raise_for_status()


@expand
class TestMissingRepo(_TestBase):
    @foreach(missing_repos)
    def test_package(self, name):
        self.assertNotIn(name, bad_metadata)
        self.assertNotIn(name, wrong_result)
        self.assertNotIn(name, sometimes_wrong_result)

        url = self._get_scm(name)
        self.assertIsNotNone(url)

        try:
            r = web_session.head(url, timeout=20)
        except Exception:
            return

        if r.status_code == 405 and "code.google.com" in url:
            r = web_session.get(url, timeout=20)
        try:
            r.raise_for_status()
        except Exception:
            pass
        else:
            if name in ["pychart"]:
                return
            self.assertTrue(False, "{} is not missing for {}".format(url, name))


@expand
class TestSetuppyMismatch(_TestBase):
    @foreach([i.strip("$") for i in setuppy_mismatches if not i.endswith("-")])
    def test_package(self, name):
        self.assertNotIn(name, bad_metadata)
        self.assertNotIn(name, wrong_result)
        self.assertNotIn(name, sometimes_wrong_result)
        self.assertNotIn(name, missing_repos)
        url = self._get_scm(name)
        self.assertIsNotNone(url)



import unittest

from pypidb._db import Database, reverse_mappings


class TestDB(unittest.TestCase):
    def test_basic(self):
        db = Database(store_fetch_list=True)
        url = db.find_project_scm_url("setuptools-scm")
        self.assertEqual(url, "https://github.com/pypa/setuptools_scm")


class TestData(unittest.TestCase):
    def test_reverse_mapping_urls(self):
        for key in reverse_mappings:
            self.assertEqual(key, key.lower())



import sys

try:
    from urllib.parse import urlparse, urlsplit, urljoin, parse_qs
except ImportError:  # pragma: no cover
    from urlparse import urlparse, urlsplit, urljoin, parse_qs

try:
    import logger_helper
except ImportError:  # pragma: no cover
    logger_helper = None

PY2 = sys.version_info[0] == 2

__all__ = ["PY2", "logger_helper", "urlparse", "urlsplit", "urljoin", "parse_qs"]



import unittest

from pypidb._exceptions import IncompletePackageMetadata, InvalidPackage

from unittest_expander import expand, foreach

from tests.data import bad_metadata, expected, failures
from tests.test_fedora import _fedora_packages
from tests.test_top import _all as _top_all
from tests.utils import _stdlib, _TestBase

try:
    from tests.datasets import get_opensuse_packages, OpenSUSEOBSLoader
except (SyntaxError, ImportError):
    raise unittest.SkipTest("Disabled on Python 2")

repeat_all = True
repeat_expected = True
repeat_failures = True


def _fetch_names(project, ignore_failures=True):
    packages = get_opensuse_packages(project)
    names = []
    for name in packages:
        if name in _stdlib:
            continue

        if not repeat_all and name.lower() in [i.lower() for i in _top_all]:
            continue

        if name.lower() in [i.lower() for i in _fedora_packages]:
            continue

        if not repeat_expected and name.lower() in [i.lower() for i in expected]:
            continue

        if not repeat_failures and name.lower() in [i.lower() for i in bad_metadata]:
            continue

        if (
            not repeat_failures
            and ignore_failures
            and name.lower() in [i.lower() for i in failures]
        ):
            continue

        names.append(name)

    return sorted(names)


@expand
class TestOpenSUSE(_TestBase):
    @foreach(OpenSUSEOBSLoader._not_pypi)
    def test_invalid_name(self, name):
        if name not in ["docs", "pyGRID"]:
            with self.assertRaises(InvalidPackage):
                self._get_scm(name)
        if not name.startswith("python-"):
            with self.assertRaises(InvalidPackage):
                self._get_scm("python-" + name)
        if not name.startswith("py") and name not in ["espeak"]:
            with self.assertRaises(InvalidPackage):
                self._get_scm("py" + name)


@expand
class TestOpenSUSEMain(_TestBase):

    names = _fetch_names(project="devel:languages:python", ignore_failures=False)
    expected_failures = _TestBase.expected_failures + [
        "ana",
        "application",
        "bencode",
        "bindep",
        "bitvector",
        "bobodoctestumentation",
        "bpython",
        "durus",
        "evtx",
        "Gloo",
        "inifile",
        "kid",
        "mysql-connector-python",
        "nagios-http-json",  # deleted
        "ovirt-engine-sdk",
        "project-config",  # invalid package
        "pychart",
        "pyjwkest",
        "pygments-ansi-color",
        "pykerberos",
        "pyprint",
        "python-axolotl-curve25519",
        "python-docs-theme",
        "pythonwhois",
        "pytidylib",
        "qet-tb-generator",
        "selection",
        "strict-rfc3339",
        "testflo",
    ]

    @foreach(names)
    def test_package(self, name):
        self._test_names([name])


@expand
class TestOpenSUSENumeric(_TestBase):

    names = _fetch_names(project="devel:languages:python:numeric")
    expected_failures = [
        "morfessor",
        "pomegranate",
        "pyfeyn",
        "pylineclip",
        "pymol",
        "pyprimes",
        "pyzo",
        "psylab",
    ]

    @foreach(names)
    def test_package(self, name):
        self._test_names([name])


@expand
class TestOpenSUSEPyTest(_TestBase):

    names = _fetch_names(project="devel:languages:python:pytest")

    @foreach(names)
    def test_package(self, name):
        self._test_names([name])


@expand
class TestOpenSUSEDjango(_TestBase):

    names = _fetch_names(project="devel:languages:python:django")

    @foreach(names)
    def test_package(self, name):
        self._test_names([name])


@expand
class TestOpenSUSEFlask(_TestBase):

    names = _fetch_names(project="devel:languages:python:flask")

    @foreach(names)
    def test_package(self, name):
        self._test_names([name])


@expand
class TestOpenSUSEJupyter(_TestBase):

    names = _fetch_names(project="devel:languages:python:jupyter")
    expected_failures = ["ipympl", "ipysheet", "jupyterlab-pygments"]

    @foreach(names)
    def test_package(self, name):
        self._test_names([name])


@expand
class TestOpenSUSEAvocado(_TestBase):

    names = _fetch_names(project="devel:languages:python:avocado")

    @foreach(names)
    def test_package(self, name):
        self._test_names([name])


@expand
class TestOpenSUSEMailman(_TestBase):

    names = _fetch_names(project="devel:languages:python:mailman")
    expected_failures = ["mailman-web"]

    @foreach(names)
    def test_package(self, name):
        self._test_names([name])


@expand
class TestOpenSUSECertbot(_TestBase):

    names = _fetch_names(project="devel:languages:python:certbot")

    @foreach(names)
    def test_package(self, name):
        self._test_names([name])


@expand
class TestOpenSUSEAzure(_TestBase):

    names = _fetch_names(project="devel:languages:python:azure")

    @foreach(names)
    def test_package(self, name):
        try:
            self._test_names([name], ignore_setuppy=True)
        except IncompletePackageMetadata:
            pass


@expand
class TestOpenSUSEAWS(_TestBase):

    names = _fetch_names("devel:languages:python:aws")

    @foreach(names)
    def test_package(self, name):
        self._test_names([name])


@expand
class TestOpenSUSEMisc(_TestBase):

    names = _fetch_names(project="devel:languages:python:misc")
    expected_failures = [
        "emencia-django-newsletter",  # missing repo
        "schevo",  # missing repo
        "schevogears",
        "schevosql",
    ]
    _allow_missing = True
    _ignore_invalid = True

    @foreach(names)
    def test_package(self, name):
        self._test_names([name])
